/*
  Using the BNO080 IMU
  By: Nathan Seidle (modified by Don Kelly)
  SparkFun Electronics
  Date: December 21st, 2017
  License: This code is public domain but you buy me a beer if you use this and we meet someday (Beerware license).

  Feel like supporting our work? Buy a board from SparkFun!
  https://www.sparkfun.com/products/14586

  This example shows how to output the parts of the calibrated gyro.

  It takes about 1ms at 400kHz I2C to read a record from the sensor, but we are polling the sensor continually
  between updates from the sensor. Use the interrupt pin on the BNO080 breakout to avoid polling.

  Important Notes:
  - Be sure to calibrate the BNO080 first so that 0.0 degrees yaw is accurately pointing North.
  - The Madgwick filter may take a few seconds to converge... give it a chance!
  
*/

#include <SPI.h>
#include <Wire.h>
#include "SparkFun_BNO080_Arduino_Library.h"
#include "globals.h"
#include "MadgwickAHRS.h"

// Create IMU object
BNO080 myIMU;

//These pins can be any GPIO
byte imuCSPin = 10;
byte imuWAKPin = 9;
byte imuINTPin = 8;
byte imuRSTPin = 7;

void setup()
{
  // Set up serial
  Serial.begin(9600);
  Serial.println();
  
  myIMU.enableDebugging(Serial); //Pipe debug messages to Serial port

  if (myIMU.beginSPI(imuCSPin, imuWAKPin, imuINTPin, imuRSTPin) == false)
  {
    Serial.println("BNO080 over SPI not detected. Are you sure you have all 6 connections? Freezing...");
    //while (1);
  }
  
  // Set up I2C
  //Wire.begin();

  // Start IMU
  //myIMU.begin();
  //Wire.setClock(400000); //Increase I2C data rate to 400kHz
  pinMode(led, OUTPUT);

  //Enable dynamic calibration for accel, gyro, and mag (may want this off)
  myIMU.calibrateAll(); //Turn on cal for Accel, Gyro, and Mag

  // Enable IMU sensors
  myIMU.enableGyro(5); //Send data update every 5ms
  //myIMU.enableLinearAccelerometer(50); //Send data update every 50ms
  myIMU.enableAccelerometer(5); //Send data update every 5ms
  myIMU.enableMagnetometer(10); //Send data update every 10ms
  myIMU.enableRotationVector(5); //Send data update every 5ms

  // Give the sensor some time to set up.
  //delay(500);
  
  // Initialize timers for dt calculation
  lastUpdate = micros();
  //lastLoopTime = micros();

}

void loop()
{
  //Look for reports from the IMU
  if (myIMU.dataAvailable() == true)
  {

    //Update delta time for estimators
    currentUpdate = micros();
    dt = ((currentUpdate - lastUpdate) * 0.000001);
    lastUpdate = currentUpdate;

    // Read IMU senor values
    float gx = myIMU.getGyroX();
    float gy = myIMU.getGyroY();
    float gz = myIMU.getGyroZ();
    
    float ax = myIMU.getAccelX();
    float ay = myIMU.getAccelY();
    float az = myIMU.getAccelZ();

    float mx = myIMU.getMagX();
    float my = myIMU.getMagY();
    float mz = myIMU.getMagZ();

    float quatI = myIMU.getQuatI();
    float quatJ = myIMU.getQuatJ();
    float quatK = myIMU.getQuatK();
    float quatReal = myIMU.getQuatReal();

    // Calculate euler angles from BNO080's quaternions (radians)   
    float q0 = quatReal; float q1 = quatI; 
    float q2 = quatJ; float q3 = quatK; 
    yaw   =  atan2((q1 * q2 + q0 * q3), ((q0 * q0 + q1 * q1) - 0.5f));   
    pitch = -asin(2.0f * (q1 * q3 - q0 * q2));
    roll  =  atan2((q0 * q1 + q2 * q3), ((q0 * q0 + q3 * q3) - 0.5f));

    // Convert BNO080 Euler angles to degrees
    yaw *= R2D;    pitch *= R2D;    roll *= R2D; 
    
    // Align BNO080 axes to NED to match Madgwick NED alignment and 
    // North to same quadrant
    pitch *= -1;
    roll *= +1;
    yaw *= -1;

    yaw += 90.0;

    // Keep BNO080 yaw between 0 and 360 degrees
    if (yaw<0){
      yaw += 360;
    }

    // Calculate quaternion from Madgwick Filter, convert to Euler angles
    MadgwickAHRSupdate(gx, -gy, -gz, ax, -ay, -az, mx, -my, -mz);        
    getYawPitchRoll(ypr);

    // Ensure Madgwick yaw is kept between 0 and 360 degrees
    yawRangeAdjust();

    // Stream Euler angles to serial
    serialPortOut();

  }
}
